import numpy as np

def blr(q, XX,tt, idxs=None):

    Ndata = XX.shape[1]

    if (idxs is None):
        idxs = np.arange(Ndata)

    X = XX[:,idxs].T
    t = tt[:,idxs].T
    alpha = 100

    # Prior
    Vprior = -0.5*np.sum(q**2) / alpha
    Fprior = -q / alpha

    # LLH
    tv = np.dot( X,q)
    exptv = np.exp(-tv)

    VV = tv*t - np.log(1+exptv) - tv

    V = np.sum( tv*t ) - np.sum( np.log( 1+exptv ) ) - np.sum(tv )
    F = X * (t - 1.0/(1.0+exptv))

    TotalV = -(V+Vprior)

    TotalF = ( Ndata*np.mean(F,0,keepdims=True)+Fprior.T ).T

    return {
    'llh' : TotalV,
    'grad' : TotalF,
    'grad_data' : F.T
    }




class SGLD():

    def __init__(self,np,ic,h,force,params):

        self.sq2h = np.sqrt(2*h)
        self.h = h

        self.np = np

        self.force = force

    def step(self, q ):

        fres = self.force(q)
        self.v,f = fres.get('llh'), fres.get('grad') 

        q = q + self.h*f  + self.sq2h*self.np.random.randn( *q.shape )

        return q

    def clear(self,q):
        pass




class LM():

    def __init__(self,np,ic,h,force,params):

        self.sqh2 = np.sqrt(h/2)
        self.h = h
        self.np = np
        self.R = np.random.randn( *ic.shape )

        self.force = force

    def step(self, q ):

        fres = self.force(q)
        self.v,f = fres.get('llh'), fres.get('grad') 

        q = q + self.h*f  + self.sqh2*self.R

        self.R = self.np.random.randn( *q.shape )

        q = q + self.sqh2*self.R

        return q

    def clear(self,q):
        pass



from __future__ import absolute_import
import numpy as np
import time

from .algos.hmc import HMC
from .algos.lm import LM
from .algos.racecar import RACECAR
from .algos.sgld import SGLD

class Sampler():

    def __init__(self,ic, h, force, algo='racecar',params={} ):

        self.timetaken = 0

        ic = np.array(ic).reshape(-1,1)
        self.q = np.copy(ic)

        if (params.get('seed') is not None):
            np.random.seed( params.get('seed') )

        if (algo.lower()=='sgld'):
            self.ig = SGLD(np,ic,h,force,params)

        if (algo.lower()=='lm'):
            self.ig = LM(np,ic,h,force,params)

        if (algo.lower()=='hmc'):
            self.ig = HMC(np,ic,h,force,params)

        if (algo.lower()=='racecar'):
            self.ig = RACECAR(np,ic,h,force,params)


    def sample(self, Nsteps, printnum=None, thinning=1):

        Qk = 0
        NQ = 1+Nsteps//thinning
        Q = np.zeros( [NQ , len(self.q.flatten() ) ] )

        self.ig.acc = 0
        self.ig.clear(self.q)

        stime = time.time()

        if (printnum is not None):
            pf = Nsteps // printnum
        else:
            pf = Nsteps+2

        for n in range(Nsteps):

            self.q = self.ig.step( self.q )

            if (n%thinning==0) and (Qk<NQ):
                Q[Qk,:] = self.q.copy().flatten()
                Qk += 1

            if (np.isnan(self.q).any() ):
                break

            if (n+1)%pf==0:
                print('Steps:',n+1 ,'  Time:',time.time() - stime ,'  V:', self.ig.v)

        self.timetaken = time.time() - stime
        Q = Q[:Qk,:]

        return Q



"""
Racecar :: A lightweight Bayesian sampling algorithm
"""
from __future__ import absolute_import

# Set explicit packages
from .sampler import Sampler as sampler
from .llh import blr

__author__ = "Charles Matthews"
__license__ = "MIT"
__maintainer__ = "Charles Matthews"
__email__ = "mail@cmatthe.ws"




class RACECAR():

    def __init__(self,np,ic,h,force,params):

        self.h = h
        self.h2 = h/2
        self.Nrk4 = params.get('Nrk4',1)
        self.rk4h = self.h2 / self.Nrk4
        self.np = np

        g = params.get('g',1)
        self.c1 = np.exp(-g*h)
        self.c3 = np.sqrt(1-self.c1*self.c1)

        self.p = np.random.randn( *ic.shape )

        NX = len(ic) * ( 1+2*params.get('Nxi',0) )
        NX = min(NX, len(ic)**2 )

        self.xi = np.random.randn( NX,1 )

        self.force = force


        fres = self.force(ic)
        self.v,self.f,ff = fres.get('llh'), fres.get('grad'), fres.get('grad_data')


        C = np.abs(np.cov( ff ))
        self.C = C.copy()

        C = C + np.eye(len(ic))*C.max()
        Cval = np.sort( C.flatten())[-NX:].min()
        self.Cx = self.C * (C>=Cval)
        C = (C >= Cval)
        self.cii,self.cjj = np.where(C)
        self.tgt = np.array(self.cii==self.cjj, dtype=int)


    def clear(self,q):
        pass

    def step(self,q):

        h = self.h
        h2 = self.h2

        self.p = self.p + h2 * self.f
        q = q + h2 * self.p

        for n in range( self.Nrk4 ):
            self.RK4()

        self.p = self.p * self.c1 + self.c3 * self.np.random.randn( *self.p.shape )

        for n in range( self.Nrk4 ):
            self.RK4()

        q = q + h2 * self.p

        fres = self.force(q)
        self.v,self.f = fres.get('llh'), fres.get('grad') 

        self.p = self.p + h2 * self.f

        return q


    def RK4(self):

        h = self.rk4h

        h2 = h / 2
        p = self.p
        xi = self.xi

        K1p, K1xi = self.df(p,xi)
        K2p, K2xi = self.df(p + h2*K1p,xi + h2*K1xi)
        K3p, K3xi = self.df(p + h2*K2p,xi + h2*K2xi)
        K4p, K4xi = self.df(p + h*K3p,xi + h*K3xi)

        pp = p + (K1p + 2*K2p + 2*K3p + K4p)*h/6
        xx = xi + (K1xi + 2*K2xi + 2*K3xi + K4xi)*h/6

        self.p = pp
        self.xi = xx


    def df(self,p,xi):

        dp = p.copy()
        self.np.add.at( dp , self.cii , -xi*( p.copy()[self.cjj] ) )
        dp = dp - p

        dxi = (p[self.cii]*p[self.cjj] - self.tgt.reshape(xi.shape) )

        return dp, dxi




class HMC():

    def __init__(self,np,ic,h,force,params):

        self.h = h
        self.h2 = h/2
        self.M = params.get('M',5)
        self.force = force
        self.np = np

        fres = self.force(ic)
        self.v,self.f = fres.get('llh'), fres.get('grad') 

        self.acc = 0

    def step(self,q):

        p = self.np.random.randn( *q.shape )

        q0 = q.copy()

        V0 = self.v
        K0 = self.np.sum( p*p )/2

        f = self.f.copy()

        h2 = self.h2
        h = self.h

        for m in range( self.M ):
            p = p + h2*f
            q = q + h*p
            fres = self.force(q)
            V1,f = fres.get('llh'), fres.get('grad')
            p = p + h2*f

        K1 = self.np.sum( p*p )/2

        dH = (V1-V0) + (K1-K0)

        if (self.np.log( self.np.random.rand() ) < -dH ):
            # Accept
            self.acc += 1
            self.v = V1
            self.f = f.copy()
            return q

        # Not accept
        return q0

    def clear(self,q):
        fres = self.force(q)
        self.v, self.f = fres.get('llh'), fres.get('grad')
        self.acc = 0
